#include "test.h"

namespace junction_test {

class JunctionMapInsDelFindTest_Parallel : public cds_test::stress_fixture {
protected:
  static unsigned s_nInsertPercentage;
  static unsigned s_nDeletePercentage;
  static size_t s_nGCFrequency; // Run GC after "s_nGCFrequency" operations.
  static size_t s_nThreadCount;
  static size_t s_nMapKeyRange;
  static size_t s_nCrudeMapCapacity;

  enum actions { do_find, do_insert, do_delete };
  static const unsigned int kShuffleSize = 100;
  static actions s_arrShuffle[kShuffleSize];

  static size_t s_nCrudePassCount;
  static size_t s_nGrampaPassCount;
  static size_t s_nLinearPassCount;
  static size_t s_nLeapfrogPassCount;

  static void InitShuffleArray() {
    // Build an array of shuffled actions.
    EXPECT_LE(s_nInsertPercentage + s_nDeletePercentage, 100);
    actions* pFirst = s_arrShuffle;
    actions* pLast = s_arrShuffle + s_nInsertPercentage;
    std::fill(pFirst, pLast, do_insert);
    pFirst = pLast;
    pLast += s_nDeletePercentage;
    std::fill(pFirst, pLast, do_delete);
    pFirst = pLast;
    pLast = s_arrShuffle + sizeof(s_arrShuffle) / sizeof(s_arrShuffle[0]);
    if (pFirst < pLast) {
      std::fill(pFirst, pLast, do_find);
    }
    std::random_device rd;
    std::mt19937 g(rd());
    std::shuffle(s_arrShuffle, pLast, g);
  }

  static void SetUpTestCase() {
    InitShuffleArray();
    const cds_test::config& cfg = get_config("ParallelJunction");
    GetConfigNonZeroExpected(InsertPercentage, 5);
    GetConfigNonZeroExpected(DeletePercentage, 5);
    GetConfigNonZeroExpected(ThreadCount, 4);
    GetConfigNonZeroExpected(MapKeyRange, 20000);
    // CrudeMap may need a big enough capacity in case may insertion operations
    // got jammed up.
    GetConfigNonZeroExpected(CrudeMapCapacity, s_nMapKeyRange * 1024);
    GetConfigNonZeroExpected(GCFrequency, 1500);
    GetConfigNonZeroExpected(CrudePassCount, 1500000000);
    GetConfigNonZeroExpected(GrampaPassCount, 650000000);
    GetConfigNonZeroExpected(LinearPassCount, 900000000);
    GetConfigNonZeroExpected(LeapfrogPassCount, 850000000);
  }

  template <typename Map>
  static void run_test(Map* map, size_t pass_count, size_t* inserted_num,
                       size_t* deleted_num) {
    auto qsbrContext = junction::DefaultQSBR.createContext();

    std::random_device rd;
    std::mt19937 gen(rd());
    std::uniform_int_distribution<size_t> dis(2, s_nMapKeyRange);

    unsigned action_index = 0;
    size_t nInsertedNum = 0;
    size_t nDeletedNum = 0;
    size_t nFindSuccess = 0;
    size_t nOperations = 0;

    for (size_t count = 0; count < pass_count; count++) {
      // The number to operate on the map.
      size_t key = dis(gen);
      switch (s_arrShuffle[action_index]) {
        case do_insert: {
          size_t val = dis(gen);
          if (map_insert(map, key, val)) {
            nInsertedNum++;
          }
          break;
        }
        case do_delete: {
          if (map_delete(map, key)) {
            nDeletedNum++;
          }
          break;
        }
        case do_find: {
          if (map_find(map, key)) {
            ++nFindSuccess;
          }
          break;
        }
        default: { break; }
      }
      if (++action_index >= kShuffleSize) {
        action_index = 0;
      }
      if (++nOperations > s_nGCFrequency) {
        junction::DefaultQSBR.update(qsbrContext);
        nOperations = 0;
      }
    }
    junction::DefaultQSBR.update(qsbrContext);
    junction::DefaultQSBR.destroyContext(qsbrContext);
    *inserted_num = nInsertedNum;
    *deleted_num = nDeletedNum;
  }

  template <typename Map>
  void JunctionThreading(Map* map, size_t pass_count) {
    std::unique_ptr<std::thread[]> threads(new std::thread[s_nThreadCount]);
    std::unique_ptr<size_t[]> inserted_nums(new size_t[s_nThreadCount]);
    std::unique_ptr<size_t[]> deleted_nums(new size_t[s_nThreadCount]);
    for (size_t i = 0; i < s_nThreadCount; i++) {
      threads[i] = std::thread(run_test<Map>, map, pass_count,
                               &inserted_nums[i], &deleted_nums[i]);
    }
    size_t inserted_sum = 0;
    size_t deleted_sum = 0;
    for (size_t i = 0; i < s_nThreadCount; i++) {
      threads[i].join();
      inserted_sum += inserted_nums[i];
      deleted_sum += deleted_nums[i];
    }
    EXPECT_LE(deleted_sum, inserted_sum);
  }
};

size_t JunctionMapInsDelFindTest_Parallel::s_nThreadCount;
size_t JunctionMapInsDelFindTest_Parallel::s_nMapKeyRange;
size_t JunctionMapInsDelFindTest_Parallel::s_nCrudeMapCapacity;
size_t JunctionMapInsDelFindTest_Parallel::s_nGCFrequency;
unsigned JunctionMapInsDelFindTest_Parallel::s_nInsertPercentage;
unsigned JunctionMapInsDelFindTest_Parallel::s_nDeletePercentage;
const unsigned int JunctionMapInsDelFindTest_Parallel::kShuffleSize;
JunctionMapInsDelFindTest_Parallel::actions JunctionMapInsDelFindTest_Parallel::
    s_arrShuffle[JunctionMapInsDelFindTest_Parallel::kShuffleSize];
size_t JunctionMapInsDelFindTest_Parallel::s_nCrudePassCount;
size_t JunctionMapInsDelFindTest_Parallel::s_nGrampaPassCount;
size_t JunctionMapInsDelFindTest_Parallel::s_nLinearPassCount;
size_t JunctionMapInsDelFindTest_Parallel::s_nLeapfrogPassCount;

TEST_F(JunctionMapInsDelFindTest_Parallel, JunctionMapCrude) {
  std::unique_ptr<CrudeMap> map(new CrudeMap(s_nCrudeMapCapacity));
  JunctionThreading(map.get(), s_nCrudePassCount);
}

TEST_F(JunctionMapInsDelFindTest_Parallel, JunctionMapLeapfrog) {
  std::unique_ptr<LeapfrogMap> map(new LeapfrogMap());
  JunctionThreading(map.get(), s_nLeapfrogPassCount);
}

TEST_F(JunctionMapInsDelFindTest_Parallel, JunctionMapLinear) {
  std::unique_ptr<LinearMap> map(new LinearMap());
  JunctionThreading(map.get(), s_nLinearPassCount);
}

TEST_F(JunctionMapInsDelFindTest_Parallel, JunctionMapGrampa) {
  std::unique_ptr<GrampaMap> map(new GrampaMap());
  JunctionThreading(map.get(), s_nGrampaPassCount);
}

} // namespace junction_test
