#include "test.h"

namespace junction_test {

class JunctionMapInsDelFindTest_Sequential : public cds_test::stress_fixture {
protected:
  static unsigned s_nInsertDeletePercentage;
  static size_t s_nGCFrequency; // Run GC after "s_nGCFrequency" operations.
  static size_t s_nMapKeyRange;
  static size_t s_nCrudeMapCapacity;

  enum actions { do_find, do_insert, do_delete };
  static const unsigned int kShuffleSize = 100;
  static actions s_arrShuffle[kShuffleSize];

  static size_t s_nCrudePassCount;
  static size_t s_nGrampaPassCount;
  static size_t s_nLinearPassCount;
  static size_t s_nLeapfrogPassCount;

  static void SetUpTestCase() {
    const cds_test::config& cfg = get_config("SequentialJunction");
    GetConfigNonZeroExpected(InsertDeletePercentage, 5);
    GetConfigNonZeroExpected(MapKeyRange, 20000);
    GetConfigNonZeroExpected(CrudeMapCapacity, s_nMapKeyRange * 64);
    GetConfigNonZeroExpected(GCFrequency, 1500);
    GetConfigNonZeroExpected(CrudePassCount, 1500000000);
    GetConfigNonZeroExpected(GrampaPassCount, 650000000);
    GetConfigNonZeroExpected(LinearPassCount, 900000000);
    GetConfigNonZeroExpected(LeapfrogPassCount, 850000000);
  }

  template <typename Map> static void run_test(Map* map, size_t pass_count) {
    auto qsbrContext = junction::DefaultQSBR.createContext();
    size_t nInsertedNum = 0;
    size_t nNotInsertedNum = 0;
    size_t nFindSuccess = 0;
    size_t nFindFailed = 0;
    size_t nDeletedNum = 0;
    size_t nOperations = 0;

    // The number to operate on the map.
    size_t n = s_nMapKeyRange;
    for (size_t count = 0; count < pass_count; count++) {
      // Insert
      unsigned mod = count % kShuffleSize;
      if (mod < s_nInsertDeletePercentage && map_insert(map, n, n)) {
        nInsertedNum++;
      } else {
        nNotInsertedNum++;
      }
      // Find
      if (map_find(map, n)) {
        ++nFindSuccess;
      } else {
        ++nFindFailed;
      }
      // Delete
      if (mod < s_nInsertDeletePercentage && map_delete(map, n)) {
        nDeletedNum++;
      }
      if (++nOperations > s_nGCFrequency) {
        junction::DefaultQSBR.update(qsbrContext);
        nOperations = 0;
      }
      if (++n == 2 * s_nMapKeyRange) {
        n = s_nMapKeyRange;
      }
    }
    junction::DefaultQSBR.update(qsbrContext);
    junction::DefaultQSBR.destroyContext(qsbrContext);
    EXPECT_EQ(nInsertedNum, nDeletedNum);
    EXPECT_EQ(nInsertedNum, nFindSuccess);
    EXPECT_EQ(nFindFailed, nNotInsertedNum);
  }
};

size_t JunctionMapInsDelFindTest_Sequential::s_nMapKeyRange;
size_t JunctionMapInsDelFindTest_Sequential::s_nCrudeMapCapacity;
size_t JunctionMapInsDelFindTest_Sequential::s_nGCFrequency;
unsigned JunctionMapInsDelFindTest_Sequential::s_nInsertDeletePercentage;
const unsigned int JunctionMapInsDelFindTest_Sequential::kShuffleSize;
JunctionMapInsDelFindTest_Sequential::actions
    JunctionMapInsDelFindTest_Sequential::s_arrShuffle
        [JunctionMapInsDelFindTest_Sequential::kShuffleSize];
size_t JunctionMapInsDelFindTest_Sequential::s_nCrudePassCount;
size_t JunctionMapInsDelFindTest_Sequential::s_nGrampaPassCount;
size_t JunctionMapInsDelFindTest_Sequential::s_nLinearPassCount;
size_t JunctionMapInsDelFindTest_Sequential::s_nLeapfrogPassCount;

TEST_F(JunctionMapInsDelFindTest_Sequential, JunctionMapCrude) {
  std::unique_ptr<CrudeMap> map(new CrudeMap(s_nCrudeMapCapacity));
  run_test(map.get(), s_nCrudePassCount);
}

TEST_F(JunctionMapInsDelFindTest_Sequential, JunctionMapLeapfrog) {
  std::unique_ptr<LeapfrogMap> map(new LeapfrogMap());
  run_test(map.get(), s_nLeapfrogPassCount);
}

TEST_F(JunctionMapInsDelFindTest_Sequential, JunctionMapLinear) {
  std::unique_ptr<LinearMap> map(new LinearMap());
  run_test(map.get(), s_nLinearPassCount);
}

TEST_F(JunctionMapInsDelFindTest_Sequential, JunctionMapGrampa) {
  std::unique_ptr<GrampaMap> map(new GrampaMap());
  run_test(map.get(), s_nGrampaPassCount);
}

} // namespace junction_test
